# Performance test results #

These tests compare prefab_classes with other modules that, as at least part of their
feature set, handle writing class boilerplate. Some of these modules may have more
features or do extra work which could make them more useful in certain situations.

The main modules being compared are prefab_classes, dataclasses, attrs and pydantic.
Some tests are also run against `namedtuple` and `NamedTuple` along with handwritten
classes.

Rough specs:
```
2023 Windows 10 Desktop: Intel i5-13600KF, Crucial P3 1TB PCIe M.2 2280 SSD
```

## Versions ##

| Software Package | Version Number |
|:---|---:|
| Python | 3.12.2 |
| prefab_classes | 0.11.0 |
| attrs | 23.2.0 |
| cattrs | 23.2.3 |
| pydantic | 2.6.4 |
| pydantic_core | 2.16.3 |



## Hyperfine tests ##

These tests are run using [Hyperfine](https://github.com/sharkdp/hyperfine)

The shell scripts to run these tests can be generated by running 
`perf/hyperfine_testmaker.py`.

### Import Time ###

This just tests the overall time to launch python and import the module to be used 
for constructing classes. `python -c "pass"` used as a baseline.

| Command | Mean [ms] | Min [ms] | Max [ms] | Relative |
|:---|---:|---:|---:|---:|
| `python -c "pass"` | 22.2 ± 0.6 | 21.1 | 24.8 | 1.00 |
| `python -c "from prefab_classes import prefab"` | 24.3 ± 0.5 | 23.6 | 26.5 | 1.09 ± 0.04 |
| `python -c "from collections import namedtuple"` | 23.7 ± 0.6 | 22.9 | 27.8 | 1.07 ± 0.04 |
| `python -c "from typing import NamedTuple"` | 31.3 ± 0.6 | 30.5 | 34.8 | 1.41 ± 0.05 |
| `python -c "from dataclasses import dataclass"` | 38.2 ± 0.6 | 37.1 | 40.8 | 1.72 ± 0.05 |
| `python -c "from attrs import define"` | 51.5 ± 0.7 | 50.5 | 54.3 | 2.32 ± 0.07 |
| `python -c "from pydantic import BaseModel"` | 68.4 ± 2.0 | 65.9 | 76.5 | 3.09 ± 0.13 |


### Class Contruction ###

This is a series of tests of the time it takes to launch python and generate 100
classes each with 5 attributes.

The code for each class looks roughly like this:

```python
@dataclass
class C0:
    a: int
    b: int
    c: int
    d: int
    e: int
```

In some cases the `__init__`, `__repr__` and `__eq__` functions are also looked up
in order to force them to be generated (prefab_eval).

For a baseline comparison, `python -c "pass"` is included just to show the overhead
from the interpreter.

As `prefab_classes` has multiple ways of generating classes it is here multiple times.

`prefab_classes_timer` generates the classes dynamically, but leaves the methods 
unused so they are not yet evaluated. This is how they would usually be on import.

`prefab_eval_timer` generates the classes dynamically, and also generates their 
methods as they would be after first use.

| Command | Mean [ms] | Min [ms] | Max [ms] | Relative |
|:---|---:|---:|---:|---:|
| `python -c "pass"` | 22.2 ± 0.8 | 20.9 | 26.2 | 1.00 |
| `python hyperfine_importers/native_classes_timer.py` | 23.1 ± 0.4 | 22.4 | 25.9 | 1.04 ± 0.04 |
| `python hyperfine_importers/namedtuples_timer.py` | 27.9 ± 0.5 | 27.0 | 29.8 | 1.26 ± 0.05 |
| `python hyperfine_importers/typed_namedtuples_timer.py` | 37.2 ± 0.6 | 36.3 | 39.8 | 1.68 ± 0.07 |
| `python hyperfine_importers/dataclasses_timer.py` | 60.0 ± 1.0 | 58.1 | 63.3 | 2.71 ± 0.11 |
| `python hyperfine_importers/attrs_timer.py` | 89.9 ± 1.1 | 87.8 | 95.1 | 4.06 ± 0.16 |
| `python hyperfine_importers/pydantic_timer.py` | 166.3 ± 3.7 | 161.1 | 176.4 | 7.51 ± 0.32 |
| `python hyperfine_importers/prefab_classes_timer.py` | 26.4 ± 1.0 | 25.5 | 32.8 | 1.19 ± 0.06 |
| `python hyperfine_importers/prefab_eval_timer.py` | 36.4 ± 0.7 | 35.3 | 38.4 | 1.64 ± 0.07 |


## json tests ##

Based on the ORJSON dataclasses performance test (but without using orjson).

With recent updates to other serialization methods, the `prefab_classes` `to_json` method
is no longer faster than the naive dataclasses as_dict by such a margin and two of the
other packages perform serialization faster.

Pydantic's conversion just uses its `.json()` method with no arguments. This has had 
a drastic improvement since the previous test and is now the fastest method. In older versions
of Pydantic this was by far the slowest method of json conversion.

`cattrs` conversion is also faster by making use of type hints to generate more direct
serialization functions.

Result from `perf/serialization/json_conversion_speed.py`

```
Python Version: 3.12.2 (tags/v3.12.2:6abddd9, Feb  6 2024, 21:26:36) [MSC v.1937 64 bit (AMD64)]
Prefab Classes version: v0.11.0
Platform: Windows-10-10.0.19045-SP0
```

| Method             | Time /s |
|--------------------|---------|
| prefab_to_json     |     0.5 |
| dataclasses_asdict |     1.0 |
| attrs_asdict       |     1.1 |
| cattrs             |     0.3 |
| pydantic           |     0.1 |

While improvements to match at least `cattrs` performance would be nice, 
the current focus is still on improving import and class generation time.

## perf_profile.py ##

This is **NOT** an overall performance test. This is just intended to check - 
very roughly - how quickly a module with a large number of generated classes 
will import.

`prefab_attributes` uses `attribute()` instead of type hints for class generation.
This test is included to make sure there haven't been regressions in either form of 
definition.

```
Python Version: 3.12.2 (tags/v3.12.2:6abddd9, Feb  6 2024, 21:26:36) [MSC v.1937 64 bit (AMD64)]
Prefab Classes version: v0.11.0
Platform: Windows-10-10.0.19045-SP0
Time for 100 imports of 100 classes defined with 5 basic attributes
```

| Method | Total Time (seconds) |
| --- | --- |
| standard classes | 0.07 |
| namedtuple | 0.30 |
| NamedTuple | 0.47 |
| dataclasses | 1.92 |
| attrs 23.2.0 | 3.53 |
| pydantic 2.6.4 | 4.12 |
| dabeaz/cluegen | 0.09 |
| dabeaz/cluegen_eval | 0.86 |
| dabeaz/dataklasses | 0.10 |
| dabeaz/dataklasses_eval | 0.10 |
| prefab v0.11.0 | 0.14 |
| prefab_attributes v0.11.0 | 0.14 |
| prefab_eval v0.11.0 | 1.04 |
